
/*
    pbrt source code Copyright(c) 1998-2010 Matt Pharr and Greg Humphreys.

    This file is part of pbrt.

    pbrt is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.  Note that the text contents of
    the book "Physically Based Rendering" are *not* licensed under the
    GNU GPL.

    pbrt is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */



#include "stdafx.h"
#include "renderers/samplerrenderer.h"
#include "scene.h"
#include "film.h"
#include "volume.h"
#include "sampler.h"
#include "integrator.h"
#include "progressreporter.h"
#include "camera.h"
#include "intersection.h"
#include <omp.h>
#include "AdaptiveSampling/AdaptiveSampling.h" // Added by Lei Xiao


static uint32_t myHash(char *key, uint32_t len)
{
    uint32_t   hash, i;
    for (hash=0, i=0; i<len; ++i) {
        hash += key[i];
        hash += (hash << 10);
        hash ^= (hash >> 6);
    }
    hash += (hash << 3);
    hash ^= (hash >> 11);
    hash += (hash << 15);
    return hash;
} 

// SamplerRendererTask Definitions
void SamplerRendererTask::Run() 
{

	//Written by Lei Xiao: when using multi-thread, GetSubSampler divides the image into multiple regions and render them separately
	//since I need to calculate the importance map for the whole image, I shouldn't do this separation.
	
	PBRT_STARTED_RENDERTASK(taskNum);
    
	handler->timer.startTimer();

	Sampler *sampler = mainSampler->GetSubSampler(taskNum, taskCount);
    if (!sampler)
     {
        reporter.Update();
        PBRT_FINISHED_RENDERTASK(taskNum);
        return;
     }

	//printf("\ninitial sampling with %d spp\n",handler->m_initSPP);


	// Get samples from _Sampler_ and update image
	// try to make it multi-thread--note that each thread should deal with different pixels!!!
	handler->setInitialSamplingNum();
			
	# pragma omp parallel
	{
	#pragma omp for schedule(dynamic) nowait

	for(int py = 0; py < handler->m_yDim; py++)
	   for(int px = 0; px < handler->m_xDim; px++)
		{
			int indPixel = py * handler->m_xDim + px;
			int sampleCount = handler->m_numMoreSamples[indPixel];
			// Declare local variables used for rendering loop
			MemoryArena arena;
			RNG rng(indPixel);
			int maxSamples = sampleCount;//this number is always 1 for random sampler
			if (!IsPowerOf2(maxSamples)) 
		    	maxSamples = RoundUpPow2(maxSamples);
			maxSamples = max(2,maxSamples);
						
			Sample *samples = origSample->Duplicate(maxSamples);
			RayDifferential *rays = new RayDifferential[maxSamples];
			Spectrum *Ls = new Spectrum[maxSamples];
			Spectrum *Ts = new Spectrum[maxSamples];
			Intersection *isects = new Intersection[maxSamples];
			
			arena.px = px;
			arena.py = py;
			arena.pind = arena.py * handler->m_xDim + arena.px;

			int NumUsedSamples = handler->getNumExistedSamples(arena.pind); // Written by Lei Xiao: for using the pre-rendered low discrepancy samples
			sampler->GetMoreSamples(samples, rng, py, px,maxSamples); // Written by Lei Xiao


			// Generate camera rays and compute radiance along rays
			for (int i = 0; i < sampleCount; ++i)  //Written by Lei Xiao: this for loop adds 'sampleCount' samples to this pixel, for the adaptive sampling, I need to change the sample number here
			{
				// Find camera ray for _sample[i]_
				PBRT_STARTED_GENERATING_CAMERA_RAY(&samples[i]);
				float rayWeight = camera->GenerateRayDifferential(samples[i], &rays[i]);

				rays[i].ScaleDifferentials(1.f / 16);
				
				PBRT_FINISHED_GENERATING_CAMERA_RAY(&samples[i], &rays[i], rayWeight);

				// Evaluate radiance along camera ray
				PBRT_STARTED_CAMERA_RAY_INTEGRATION(&rays[i], &samples[i]);
				if (visualizeObjectIds) 
				{
					if (rayWeight > 0.f && scene->Intersect(rays[i], &isects[i])) 
					{
						// random shading based on shape id...
						uint32_t ids[2] = { isects[i].shapeId, isects[i].primitiveId };
						uint32_t h = myHash((char *)ids, sizeof(ids));
						float rgb[3] = { (h & 0xff), (h >> 8) & 0xff, (h >> 16) & 0xff };
						Ls[i] = Spectrum::FromRGB(rgb);
						Ls[i] /= 255.f;
					}
					else
						Ls[i] = 0.f;
				}
				else 
				{
					if (rayWeight > 0.f)
						Ls[i] = rayWeight * renderer->Li(scene, rays[i], &samples[i], rng,
													 arena, &isects[i], &Ts[i]);
					else 
					  {
						Ls[i] = 0.f;
						Ts[i] = 1.f;
					  }

					// Issue warning if unexpected radiance value returned
					if (Ls[i].HasNaNs()) 
					{
					   Error("Not-a-number radiance value returned "
						  "for image sample.  Setting to black.");
					   Ls[i] = Spectrum(0.f);
					}
					else if (Ls[i].y() < -1e-5) 
					{
					   Error("Negative luminance value, %f, returned"
						  "for image sample.  Setting to black.", Ls[i].y());
					   Ls[i] = Spectrum(0.f);
					}
					else if (isinf(Ls[i].y())) 
					{
					  Error("Infinite luminance value returned"
						  "for image sample.  Setting to black.");
					  Ls[i] = Spectrum(0.f);
					}
				}
				
				float rgb[3];
				Ls[i].ToRGB(rgb);
				handler->addFinalColor(rgb[0],rgb[1],rgb[2],arena.pind);


				PBRT_FINISHED_CAMERA_RAY_INTEGRATION(&rays[i], &samples[i], &Ls[i]);
			}
			

			// Written by Lei Xiao: the following several lines of code adds the rendered samples to the 'file' class
			// Report sample results to _Sampler_, add contributions to image
			if (sampler->ReportResults(samples, rays, Ls, isects, sampleCount))
			{
				for (int i = 0; i < sampleCount; ++i)
				{
					PBRT_STARTED_ADDING_IMAGE_SAMPLE(&samples[i], &rays[i], &Ls[i], &Ts[i]);
					camera->film->AddSample(samples[i], Ls[i]);
					PBRT_FINISHED_ADDING_IMAGE_SAMPLE();
				}
			}

			// Free _MemoryArena_ memory from computing image sample values
			arena.FreeAll();
			delete[] samples;
			delete[] rays;
			delete[] Ls;
			delete[] Ts;
			delete[] isects;
		}
	}
	
	float importanceT = 0;

	if(1)
	{
		//start adaptive sampling process
		//printf("adaptive sampling...\n");
		for(int it = 1; it < handler->m_numIteration; it++)
		{

			// compute the importance map based on sampled image from last iteration.
			//printf("-----computing the importance map...\n");
			
			myTimer importanceTime;
			importanceTime.startTimer();
			handler->calImportanceMap(it);
			importanceTime.stopTimer();
			importanceT += importanceTime.getTime();
			
			
			int _maxSamples = 0; //sampler->MaximumSampleCount();//this number is always 1 for PBRT's original random sampler // Written by Lei Xiao
			for(int pp=0; pp<handler->m_yDim*handler->m_xDim; pp++)
				_maxSamples = max(_maxSamples, handler->m_numMoreSamples[pp]);
			if (!IsPowerOf2(_maxSamples)) 
				_maxSamples = RoundUpPow2(_maxSamples);
			_maxSamples = max(2,_maxSamples);

			//printf("-----rendering...\n");

			sampler = mainSampler->GetSubSampler(taskNum, taskCount);

			if (!sampler)
			{
				reporter.Update();
				PBRT_FINISHED_RENDERTASK(taskNum);
				return;
			}

		#pragma omp parallel
		{
		#pragma omp for schedule(dynamic) nowait

		for(int py = 0; py < handler->m_yDim; py++)
			for(int px = 0; px < handler->m_xDim; px++)
  			{
    			int indPixel = py * handler->m_xDim + px;
				int sampleCount = handler->m_numMoreSamples[indPixel];
				assert(sampleCount<=_maxSamples);
				sampleCount = max(0,sampleCount);
				if(sampleCount == 0)
					continue;
				
				// Declare local variables used for rendering loop
				MemoryArena arena;
				RNG rng(indPixel);
				//int maxSamples = RoundUpPow2(sampleCount);
				
				Sample *samples = origSample->Duplicate(_maxSamples);
				RayDifferential *rays = new RayDifferential[_maxSamples];
				Spectrum *Ls = new Spectrum[_maxSamples];
				Spectrum *Ts = new Spectrum[_maxSamples];
				Intersection *isects = new Intersection[_maxSamples];
				
				arena.px = px;
				arena.py = py;
				arena.pind = arena.py * handler->m_xDim + arena.px;
				
				int NumUsedSamples = handler->getNumExistedSamples(arena.pind); // Written by Lei Xiao: for using the pre-rendered low discrepancy samples
				sampler->GetMoreSamples(samples, rng,py,px,_maxSamples);

				// Generate camera rays and compute radiance along rays
				for (int i = 0; i < sampleCount; ++i)  //Written by Lei Xiao: this for loop adds 'sampleCount' samples to this pixel, for my adaptive sampling, I need to change the sample number here
				{
					// Find camera ray for _sample[i]_
					PBRT_STARTED_GENERATING_CAMERA_RAY(&samples[i]);
					float rayWeight = camera->GenerateRayDifferential(samples[i], &rays[i]);

					rays[i].ScaleDifferentials(1.f /16);// Modified by Lei Xiao to magnify the bump-map effect

					PBRT_FINISHED_GENERATING_CAMERA_RAY(&samples[i], &rays[i], rayWeight);

					// Evaluate radiance along camera ray
					PBRT_STARTED_CAMERA_RAY_INTEGRATION(&rays[i], &samples[i]);
					if (visualizeObjectIds) 
					{
						if (rayWeight > 0.f && scene->Intersect(rays[i], &isects[i])) 
						{
							// random shading based on shape id...
							uint32_t ids[2] = { isects[i].shapeId, isects[i].primitiveId };
							uint32_t h = myHash((char *)ids, sizeof(ids));
							float rgb[3] = { (h & 0xff), (h >> 8) & 0xff, (h >> 16) & 0xff };
							Ls[i] = Spectrum::FromRGB(rgb);
							Ls[i] /= 255.f;
						}
						else
							Ls[i] = 0.f;
					}
					else 
					{
						if (rayWeight > 0.f)
							Ls[i] = rayWeight * renderer->Li(scene, rays[i], &samples[i], rng,
							arena, &isects[i], &Ts[i]);
						else 
						{
							Ls[i] = 0.f;
							Ts[i] = 1.f;
						}

						// Issue warning if unexpected radiance value returned
						if (Ls[i].HasNaNs()) 
						{
							Error("Not-a-number radiance value returned "
								"for image sample.  Setting to black.");
							Ls[i] = Spectrum(0.f);
						}
						else if (Ls[i].y() < -1e-5) 
						{
							Error("Negative luminance value, %f, returned"
								"for image sample.  Setting to black.", Ls[i].y());
							Ls[i] = Spectrum(0.f);
						}
						else if (isinf(Ls[i].y())) 
						{
							Error("Infinite luminance value returned"
								"for image sample.  Setting to black.");
							Ls[i] = Spectrum(0.f);
						}
					}

					
					float rgb[3];
					Ls[i].ToRGB(rgb);
					handler->addFinalColor(rgb[0],rgb[1],rgb[2],arena.pind);
					
					// Added by Lei Xiao

					PBRT_FINISHED_CAMERA_RAY_INTEGRATION(&rays[i], &samples[i], &Ls[i]);
				}

				// Written by Lei Xiao: the following several lines of code adds the rendered samples to the 'file' class
				// Report sample results to _Sampler_, add contributions to image
				if (sampler->ReportResults(samples, rays, Ls, isects, sampleCount))
				{
					for (int i = 0; i < sampleCount; ++i)
					{
						PBRT_STARTED_ADDING_IMAGE_SAMPLE(&samples[i], &rays[i], &Ls[i], &Ts[i]);
						camera->film->AddSample(samples[i], Ls[i]);
						PBRT_FINISHED_ADDING_IMAGE_SAMPLE();
					}
				}

				// Free _MemoryArena_ memory from computing image sample values
				arena.FreeAll();
				delete[] samples;
				delete[] rays;
				delete[] Ls;
				delete[] Ts;
				delete[] isects;
		   }
		   }

			if (it == handler->m_numIteration - 1)
			{
				handler->imWriteAdaptiveSampledImage();
				handler->imWriteSampleDistribution();
			}

		}
	}

	//Written by Lei Xiao: now all the rendering is finished. the following several lines of code cleans up the pointers
	//Clean up after _SamplerRendererTask_ is done with its image region
	camera->film->UpdateDisplay(sampler->xPixelStart,
		sampler->yPixelStart, sampler->xPixelEnd+1, sampler->yPixelEnd+1);
	delete sampler;
	reporter.Update();

    handler->timer.stopTimer();
	float renderTime = handler->timer.getTime();
	    
   
	float cal_time = handler->noiseMap(handler->m_numIteration);
	

	
	PBRT_FINISHED_RENDERTASK(taskNum);
	

	// save the time cost
	char fn[300];
	sprintf(fn,"%s\\TimingBreakdown.txt",handler->m_path);
	FILE* thfile = fopen(fn,"a+");
	fprintf(thfile,"Calculating importance map = %.1f seconds\n", importanceT);
	fprintf(thfile,"Rendering samples = %.1f seconds\n", renderTime - importanceT);
	fprintf(thfile,"---------------------------------------------------\n");
	fprintf(thfile,"Adaptive sampling = %.1f seconds\n\n", renderTime);

	fprintf(thfile,"Calculating noise map = %.1f seconds\n", cal_time);
	
	fclose(thfile);

}



// SamplerRenderer Method Definitions
SamplerRenderer::SamplerRenderer(Sampler *s, Camera *c,
                                 SurfaceIntegrator *si, VolumeIntegrator *vi,
                                 bool visIds) {
    sampler = s;
    camera = c;
    surfaceIntegrator = si;
    volumeIntegrator = vi;
    visualizeObjectIds = visIds;
}


SamplerRenderer::~SamplerRenderer() {
    delete sampler;
    delete camera;
    delete surfaceIntegrator;
    delete volumeIntegrator;
}


void SamplerRenderer::Render(const Scene *scene) {

	
    PBRT_FINISHED_PARSING();
    // Allow integrators to do preprocessing for the scene
    PBRT_STARTED_PREPROCESSING();
    surfaceIntegrator->Preprocess(scene, camera, this);
    volumeIntegrator->Preprocess(scene, camera, this);
    PBRT_FINISHED_PREPROCESSING();
    PBRT_STARTED_RENDERING();
    // Allocate and initialize _sample_
    Sample *sample = new Sample(sampler, surfaceIntegrator,
                                volumeIntegrator, scene);
	// Added by Nima
	handler->SetScaleAndGamma(camera->film->m_scale, camera->film->m_gamma);

    // Create and launch _SamplerRendererTask_s for rendering image

    // Compute number of _SamplerRendererTask_s to create for rendering
    int nPixels = camera->film->xResolution * camera->film->yResolution;
    int nTasks = 1; // Written by Lei Xiao
    nTasks = RoundUpPow2(nTasks);



	
	handler->allocateMemory(camera->film->xResolution,camera->film->yResolution,sampler->samplesPerPixel); // Written by Lei Xiao
	
    
	ProgressReporter reporter(nTasks, "Rendering");
    vector<Task *> renderTasks;
    for (int i = 0; i < nTasks; ++i)
        renderTasks.push_back(new SamplerRendererTask(scene, this, camera,
                                                      reporter, sampler, sample, 
                                                      visualizeObjectIds, 
                                                      nTasks-1-i, nTasks));
    EnqueueTasks(renderTasks);
    WaitForAllTasks();
    for (uint32_t i = 0; i < renderTasks.size(); ++i)
        delete renderTasks[i];
    reporter.Done();
    PBRT_FINISHED_RENDERING();
    // Clean up after rendering and store final image
    delete sample;
    //camera->film->WriteImage();
}


Spectrum SamplerRenderer::Li(const Scene *scene,
        const RayDifferential &ray, const Sample *sample, RNG &rng,
        MemoryArena &arena, Intersection *isect, Spectrum *T) const {
    Assert(ray.time == sample->time);
    Assert(!ray.HasNaNs());
    // Allocate local variables for _isect_ and _T_ if needed
    Spectrum localT;
    if (!T) T = &localT;
    Intersection localIsect;
    if (!isect) isect = &localIsect;
    Spectrum Li = 0.f;
    if (scene->Intersect(ray, isect))
	{
        Li = surfaceIntegrator->Li(scene, this, ray, *isect, sample,
                                   rng, arena);
		// Written by Lei Xiao
		BSDF *bsdf = isect->GetBSDF(ray, arena);
		Normal n = bsdf->dgShading.nn;
		Point p = bsdf->dgShading.p;
	}
    else {
        // Handle ray that doesn't intersect any geometry
        for (uint32_t i = 0; i < scene->lights.size(); ++i)
           Li += scene->lights[i]->Le(ray);
    }
    Spectrum Lvi = volumeIntegrator->Li(scene, this, ray, sample, rng,
                                        T, arena);
    return *T * Li + Lvi;
}


Spectrum SamplerRenderer::Transmittance(const Scene *scene,
        const RayDifferential &ray, const Sample *sample, RNG &rng,
        MemoryArena &arena) const {
    return volumeIntegrator->Transmittance(scene, this, ray, sample,
                                           rng, arena);
}


