
/*
    pbrt source code Copyright(c) 1998-2010 Matt Pharr and Greg Humphreys.

    This file is part of pbrt.

    pbrt is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.  Note that the text contents of
    the book "Physically Based Rendering" are *not* licensed under the
    GNU GPL.

    pbrt is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */


// samplers/random.cpp*
#include "stdafx.h"
#include "AdaptiveSampling\globals.h"
#include "samplers/random.h"
#include "montecarlo.h"
#include "camera.h"
#include "AdaptiveSampling/AdaptiveSampling.h" // Added by Lei Xiao
RandomSampler::RandomSampler(int xstart, int xend,
        int ystart, int yend, int ns, float sopen, float sclose)
    : Sampler(xstart, xend, ystart, yend, ns, sopen, sclose) {
    xPos = xPixelStart;
    yPos = yPixelStart;
    
	nSamples = ns; 
    
	// Written by Lei Xiao
	int maxSamples = handler->m_maxNumSPP;
	// Get storage for a pixel's worth of stratified samples
	imageSamples = AllocAligned<float>(5 * maxSamples);
	lensSamples = imageSamples + 2 * maxSamples;
	timeSamples = lensSamples + 2 * maxSamples;

	RNG rng(xstart + ystart * (xend-xstart));
	for (int i = 0; i < 5 * maxSamples; ++i)
		imageSamples[i] = rng.RandomFloat();

	// Shift image samples to pixel coordinates
	for (int o = 0; o < 2 * maxSamples; o += 2) {
		imageSamples[o]   += xPos;
		imageSamples[o+1] += yPos;
	}
	samplePos = 0;
	// End of written by Lei Xiao
}



Sampler *RandomSampler::GetSubSampler(int num, int count) {
    int x0, x1, y0, y1;
    ComputeSubWindow(num, count, &x0, &x1, &y0, &y1);
    if (x0 == x1 || y0 == y1) return NULL;
    return new RandomSampler(x0, x1, y0, y1, nSamples,
       shutterOpen, shutterClose);
}



// Written by Lei Xiao: it seems I need to change this function to control the x-y pixel-position
// this function is used for the initial sampling in my adaptive sampling process
int RandomSampler::GetMoreSamples(Sample *sample, RNG &rng) {
    if (samplePos == nSamples) 
	{
        if (xPixelStart == xPixelEnd || yPixelStart == yPixelEnd)
            return 0;
        if (++xPos == xPixelEnd) {
            xPos = xPixelStart;
            ++yPos;
        }
        if (yPos == yPixelEnd)
            return 0;

        for (int i = 0; i < 5 * handler->m_maxNumSPP; ++i)
            imageSamples[i] = rng.RandomFloat(); //Comment by Lei Xiao: this means that for each pixel, the imageSamples would be different

        // Shift image samples to pixel coordinates
        for (int o = 0; o < 2 * handler->m_maxNumSPP; o += 2) 
		{
            imageSamples[o]   += xPos;
            imageSamples[o+1] += yPos;
        }
        samplePos = 0;
    }
    // Return next \mono{RandomSampler} sample point

    sample->imageX = imageSamples[2*samplePos];
    sample->imageY = imageSamples[2*samplePos+1];
    sample->lensU = lensSamples[2*samplePos];
    sample->lensV = lensSamples[2*samplePos+1];
	    
	sample->time = Lerp(timeSamples[samplePos], shutterOpen, shutterClose);
    
	// Generate stratified samples for integrators
    for (uint32_t i = 0; i < sample->n1D.size(); ++i)
        for (uint32_t j = 0; j < sample->n1D[i]; ++j)
            sample->oneD[i][j] = rng.RandomFloat();
    for (uint32_t i = 0; i < sample->n2D.size(); ++i)
        for (uint32_t j = 0; j < 2*sample->n2D[i]; ++j)
            sample->twoD[i][j] = rng.RandomFloat();

    ++samplePos;
    return 1;
}

// Written by Lei Xiao: this following added function is for multi-thread only
void RandomSampler::GetMoreSamples(Sample *samples, RNG &rng, int py, int px,int sampleCount)
{

	float* myImageSamples, *myLensSamples, *myTimeSamples;
	myImageSamples = AllocAligned<float>(5 * sampleCount);
	myLensSamples = myImageSamples + 2 * sampleCount;
	myTimeSamples = myImageSamples + 2 * sampleCount;
	for (int i = 0; i < 5*sampleCount; ++i)
		myImageSamples[i] = rng.RandomFloat();

	// Shift image samples to pixel coordinates
	for (int o = 0; o < 2 * sampleCount; o += 2) 
	{
		myImageSamples[o]   += px;
		myImageSamples[o+1] += py;
	}

	for(int s=0;s<sampleCount;s++)
	{
		samples[s].imageX = myImageSamples[2*s];
		samples[s].imageY = myImageSamples[2*s+1];
		samples[s].lensU = myLensSamples[2*s];
		samples[s].lensV = myLensSamples[2*s+1];
		samples[s].time = Lerp(myTimeSamples[s], shutterOpen, shutterClose);

		// Generate stratified samples for integrators
		for (uint32_t i = 0; i < samples[s].n1D.size(); ++i)
			for (uint32_t j = 0; j <samples[s].n1D[i]; ++j)
				samples[s].oneD[i][j] = rng.RandomFloat();
		for (uint32_t i = 0; i < samples[s].n2D.size(); ++i)
			for (uint32_t j = 0; j < 2*samples[s].n2D[i]; ++j)
				samples[s].twoD[i][j] = rng.RandomFloat();

	}

 FreeAligned(myImageSamples);


}


//Written by Lei Xiao: for using the pre-rendered low discrepancy samples
void RandomSampler::GetMoreSamples(Sample *samples, RNG &rng, int py, int px,int NumUsedSamples, int sampleCount)
{

	float* myImageSamples, *myLensSamples, *myTimeSamples;
	myImageSamples = AllocAligned<float>(5 * sampleCount);
	myLensSamples = myImageSamples + 2 * sampleCount;
	myTimeSamples = myImageSamples + 2 * sampleCount;
	for (int i = 0; i < 5*sampleCount; ++i)
		myImageSamples[i] = rng.RandomFloat();

	// Shift image samples to pixel coordinates
	for (int o = 0; o < 2 * sampleCount; o += 2) 
	{
		myImageSamples[o]   += px;
		myImageSamples[o+1] += py;
	}

	for(int s=0;s<sampleCount;s++)
	{
		//if(handler->m_isRandom)
		{
			samples[s].imageX = myImageSamples[2*s]; 
			samples[s].imageY = myImageSamples[2*s+1];
			samples[s].lensU = myLensSamples[2*s];
			samples[s].lensV = myLensSamples[2*s+1];
			samples[s].time = Lerp(myTimeSamples[s], shutterOpen, shutterClose);
		}
		


		// Generate stratified samples for integrators
		for (uint32_t i = 0; i < samples[s].n1D.size(); ++i)
			for (uint32_t j = 0; j <samples[s].n1D[i]; ++j)
				samples[s].oneD[i][j] = rng.RandomFloat();
		for (uint32_t i = 0; i < samples[s].n2D.size(); ++i)
			for (uint32_t j = 0; j < 2*samples[s].n2D[i]; ++j)
				samples[s].twoD[i][j] = rng.RandomFloat();

	}

	FreeAligned(myImageSamples);


}

// Written by Lei Xiao: this following added function is for single-thread only
bool RandomSampler::GetMoreSamples(Sample *samples, RNG &rng, const int* m_numMoreSamples, int &sampleCount, int xdim) // Modified by Lei Xiao
{ 	
	int indpixel = yPos*xdim+xPos;
	sampleCount = m_numMoreSamples[indpixel];
	if (yPos == yPixelEnd)
		return 0;
	
	for (int i = 0; i < 5 * handler->m_maxNumSPP; ++i)
		imageSamples[i] = rng.RandomFloat();

	// Shift image samples to pixel coordinates
	for (int o = 0; o < 2 * handler->m_maxNumSPP; o += 2) 
	{
		imageSamples[o]   += xPos;
		imageSamples[o+1] += yPos;
	}
		
	// Return next \mono{RandomSampler} sample point
    for(int s=0;s<sampleCount;s++)
	{
  	  samples[s].imageX = imageSamples[2*s];
	  samples[s].imageY = imageSamples[2*s+1];
	  samples[s].lensU = lensSamples[2*s];
	  samples[s].lensV = lensSamples[2*s+1];
      samples[s].time = Lerp(timeSamples[s], shutterOpen, shutterClose);
 
      // Generate stratified samples for integrators
	  for (uint32_t i = 0; i < samples[s].n1D.size(); ++i)
		for (uint32_t j = 0; j <samples[s].n1D[i]; ++j)
			samples[s].oneD[i][j] = rng.RandomFloat();
	  for (uint32_t i = 0; i < samples[s].n2D.size(); ++i)
		for (uint32_t j = 0; j < 2*samples[s].n2D[i]; ++j)
			samples[s].twoD[i][j] = rng.RandomFloat();
	
	}

	if (++xPos == xPixelEnd) 
	{
		xPos = xPixelStart;
		++yPos;
	}

	return 1;
}



Sampler *CreateRandomSampler(const ParamSet &params,
                       const Film *film, const Camera *camera) {
    int ns = params.FindOneInt("pixelsamples", 4);
    int xstart, xend, ystart, yend;
    film->GetSampleExtent(&xstart, &xend, &ystart, &yend);
    return new RandomSampler(xstart, xend, ystart, yend, ns,
                             camera->shutterOpen, camera->shutterClose);
}


